#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/Vote.pm,v 1.14 2005/06/10 16:17:17 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::Vote>> - Aranea module for performing candidate
voting

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module collates the answer candidates by voting.  Each
candidate's score is set to the number of times that candidate appears
in the C<candidate_list>.  In effect, duplicate candidates are
compressed into a single entry to a high score.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::Vote;

use strict;
use warnings;
use lib qw(lib pm);                         # needed for command-line mode

use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::XML::CandidateEntry;

#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: Vote.pm,v 1.14 2005/06/10 16:17:17 jimmylin Exp $ ');

    # if there are no requests, then there is nothing to do
    return $Aranea if $Aranea->request_list()->entries_count() == -1;

    my $entries = $Aranea->candidate_list()->entries();

    # to keep track of unique candidates
    my %candidate_hash=();

    # interate through all candidates and tally up the scores.  BTW,
    # DProf shows that `foreach' is faster than `map'
    foreach my $entry (@$entries) {
	$candidate_hash{$entry->candidate()} += $entry->score();
    }

    my $new_entries = [];
    for my $element ( keys %candidate_hash ) {
	next unless $candidate_hash{$element} > 2;
	my $new_entry = Aranea::XML::CandidateEntry
	    ->new($element, $candidate_hash{$element});

	push @$new_entries, $new_entry
    }

    $Aranea->candidate_list->entries($new_entries);

    return $Aranea;
}
#>

1;
